"""
Utilities for generating a range of learning rate schedules.
"""

import math
from typing import Optional, List, Union, Callable


def ascent(step: int, total_steps: int) -> float:
    """
    Get a sequence of numbers evenly spaced between 0 and 1 so that the first
        number is 0 and the last is 1 and there are `total_steps` numbers in
        the sequence.
    """
    return round(step / (total_steps - 0.999), 8)


def triangle(step: int, total_steps: int) -> float:
    """
    Get a triangular sequence of numbers between 0 and 1, going up in half of
        `total_steps` and coming down in the other half, peaking at ~1.
    """
    half = int(math.ceil(total_steps / 2))
    if step < half:
        return 2 * ascent(step, total_steps)
    else:
        return 2 - 2 * ascent(step, total_steps)


def cosine(step: int, total_steps: int) -> float:
    """
    Get a sequence of numbers between 0 and 1 in the shape of a cosine wave with
        wavelength `total_steps`.
    """
    if total_steps < 1:
        raise ValueError(f"total_steps must be >= 1, got {total_steps}")
    angle = (step / (total_steps - 0.999)) * (2 * math.pi)
    return round((math.cos(angle) + 1) / 2, 8)


def quarter_circle(step: int, total_steps: int) -> float:
    """
    Get a sequence of numbers between 0 and 1 in the shape of a quarter-circle with
        radius `total_steps'.
    """
    if total_steps < 1:
        raise ValueError(f"total_steps must be >= 1, got {total_steps}")
    x = 0 if total_steps == 1 else step / (total_steps - 1)
    # x^2 + y^2 = r^2 = 1
    # Therefore y^2 = 1 - x^2
    # Therefore y^2 = (1 + x)(1 - x)
    y_squared = max(1 - x, 0) * (1 + x)
    return math.sqrt(y_squared)


def half_cosine(step: int, total_steps: int) -> float:
    """
    Get a sequence of numbers between 0 and 1 in the shape of the descending
        half of a cosine wave with wavelength 2*`total_steps`.
    """
    return cosine(step, (total_steps * 2) - 1)


def cycloid(step: int, total_steps: int) -> float:
    """
    Get a sequence of numbers between 0 and 1 in the shape of a cycloid with
        circle diameter 1.0 and `total_steps/(2*math.pi)` steps per cycle.
    """
    x = step * (math.pi / (total_steps - 1))

    def fx(t):
        return 0.5 * (t - math.sin(t)) - x

    def fx_prime(t):
        return 0.5 - 0.5 * math.cos(t)

    def fy_prime(t):
        return 0.5 - 0.5 * -math.sin(t)

    angle_estimate = 0.5 * x

    # XXX: 200 iterations is too many! Use a more efficient root finding algorithm
    for _ in range(200):
        if abs(fx_prime(angle_estimate)) > 0.1:
            update = fx(angle_estimate) / fx_prime(angle_estimate)
        else:
            update = fx(angle_estimate) / fy_prime(angle_estimate)
        angle_estimate = angle_estimate - update

    return 0.5 * (1 - math.cos(angle_estimate))


def half_cycloid(step: int, total_steps: int) -> float:
    return cycloid(total_steps + step, 2 * total_steps)


FN_LIBRARY = {
    "ascent": ascent,
    "triangle": triangle,
    "cosine": cosine,
    "half_cosine": half_cosine,
    "quarter_circle": quarter_circle,
    "half_cycloid": half_cycloid,
}


class Cycle:
    def __init__(
        self,
        generating_function: Union[str, Callable],
        training_examples,
        epochs,
        batch_size,
        t_0: Optional[int] = None,
        t_mult: float = 1.0,
        t_scale: float = 1.0,
        low=0.0,
        high=1.0,
        reflect=False,
    ):
        self.training_examples = training_examples
        self.epochs = epochs
        self.batch_size = batch_size
        self.total_steps = int(
            epochs * (math.floor(training_examples / batch_size) + 1)
        )

        self.t_0 = (
            t_0 * (training_examples / batch_size)
            if t_0 is not None
            else self.total_steps
        )
        self.t_mult = t_mult
        self.t_scale = t_scale

        self.low = low
        self.high = high

        self.reflect = reflect

        if not callable(generating_function):
            if generating_function in FN_LIBRARY:
                self._generating_function = FN_LIBRARY[generating_function]
            else:
                raise NotImplementedError(
                    "`generating_function` must be a callable object or one of "
                    '"ascent", "triangle", "cosine", "half_cosine", "quarter_circle" '
                    'or "half_cycloid"'
                )
        else:
            self._generating_function = generating_function

    def _get_window(self, step):
        windows = self._windows()
        cumulative = [
            sum([w[0] for w in windows][: i + 1]) for i in range(len(windows))
        ]
        position = None
        local_step = None
        for i, c in enumerate(cumulative):
            if c > step:
                position = i
                local_step = step if i == 0 else step - cumulative[i - 1]
                break
        window_width, window_height = windows[position]
        return window_width, local_step, window_height

    def _generate(self, step) -> list:
        total_steps, step, scale = self._get_window(step)
        y = self._generating_function(step, total_steps)
        y = y * scale
        y = 1 - y if self.reflect else y
        return y * (self.high - self.low) + self.low

    def __call__(self, n):
        return self._generate(n)

    def __len__(self):
        return self.total_steps

    def _windows(self):
        assert self.t_mult > 0

        # Get tile widths
        widths = [self.t_0]
        while True:
            next_item = widths[-1] * self.t_mult
            if sum(widths) + next_item <= self.total_steps:
                widths.append(next_item)
            else:
                break
        for i in range(1, len(widths)):
            widths[i] = int(widths[i] * (self.total_steps / sum(widths)))
        widths[-1] += self.total_steps - sum(widths)

        # Get tile heights
        heights = [1.0 * self.t_scale**i for i in range(len(widths))]

        return list(zip(widths, heights, strict=True))

    @property
    def stats(self) -> float:
        """
        Returns the area (as a percentage of the area of a curve where the learning
            rate is constant max_lr), percentage ascent steps and percentage descent
            steps of a learning rate schedule.
        """
        total_area = 0
        max_area = 0
        ascent_steps = 0
        descent_steps = 0
        avg_up_gradient = 0
        avg_down_gradient = 0
        total_gradient = 0
        previous_lr = None
        for s in range(self.total_steps):
            height = self(s)
            total_area += height
            max_area += 1
            if previous_lr is None:
                pass
            elif previous_lr > height:
                descent_steps += 1
                avg_down_gradient += height - previous_lr
                total_gradient += height - previous_lr
            elif previous_lr < height:
                ascent_steps += 1
                avg_up_gradient += height - previous_lr
                total_gradient += height - previous_lr
            else:
                total_gradient += height
            previous_lr = height
        return {
            "area": total_area / max_area,
            "pc_ascent": round(ascent_steps / self.total_steps, 3),
            "pc_descent": round(descent_steps / self.total_steps, 3),
            "avg_up_gradient": round(avg_up_gradient, 3),
            "avg_down_gradient": round(avg_down_gradient, 3),
            "avg_gradient": round(-(self.high - self.low) / self.total_steps, 3),
        }


class CycleProduct(Cycle):
    def __init__(self, cycles: List[Cycle], reflect=False, normalise: bool = False):
        """
        Args:
            normalise: if true, the square root of the product is returned (i.e.
                the geometric mean of the two cycles that were multiplied together)
        """
        main_training_examples = cycles[0].training_examples
        main_batch_size = cycles[0].batch_size

        assert all(c.training_examples == main_training_examples for c in cycles)
        assert all(c.batch_size == main_batch_size for c in cycles)

        self.cycles = cycles
        self.reflect = reflect
        self.normalise = normalise

        def generating_function(step: int, total_steps: int) -> float:
            output = self.cycles[0](step)
            for c in self.cycles[1:]:
                output *= c(step % c.total_steps)
            if self.normalise:
                output = math.sqrt(output)
            return output

        super().__init__(
            generating_function=generating_function,
            training_examples=self.cycles[0].training_examples,
            epochs=self.cycles[0].epochs,
            batch_size=self.cycles[0].batch_size,
            reflect=reflect,
        )


class CycleSequence:
    def __init__(self, cycles: List[Cycle]):
        self.total_steps = sum([c.total_steps for c in cycles])
        self.cycles = cycles

    def _generate(self, step):
        cycle, step = self._get_cycle_and_step(step)
        return self.cycles[cycle](step)

    def _get_cycle_and_step(self, step):
        cycle_lengths = [c.total_steps for c in self.cycles]
        cumulative = [sum(cycle_lengths[: i + 1]) for i in range(len(cycle_lengths))]
        cycle = None
        local_step = None
        for i, c in enumerate(cumulative):
            if c > step:
                cycle = i
                local_step = step if i == 0 else step - cumulative[i - 1]
                break
        return cycle, local_step

    def __call__(self, step):
        return self._generate(step)

    def __len__(self):
        return self.total_steps

    @property
    def stats(self) -> float:
        """
        Returns the area (as a percentage of the area of a curve where the learning
            rate is constant max_lr), percentage ascent steps and percentage descent
            steps of a learning rate schedule.
        """
        cycle_ratios = [c.total_steps for c in self.cycles]
        cycle_stats = {k: v * cycle_ratios[0] for k, v in self.cycles[0].stats.items()}
        for i, cycle in enumerate(self.cycles):
            if i == 0:
                continue  # We already did the first one, above
            for k, v in cycle.stats.items():
                cycle_stats[k] += cycle_ratios[i] * v

        return {k: v / self.total_steps for k, v in cycle_stats.items()}



import math
import warnings
from collections import defaultdict

from torch.optim import AdamW

EXCLUDE_FROM_WEIGHT_DECAY = ["nondecay", "bias", "norm", "embedding", "beta"]


def register_optimiser_recursive(module, optimizer):
    """
    Recursively walks the module tree and registers the optimiser
        with any layer that supports .register_optimiser()
    """
    # Check if the module itself has the method
    if hasattr(module, "register_optimiser") and callable(module.register_optimiser):
        module.register_optimiser(optimizer)

    # Recursively check children
    for child in module.children():
        register_optimiser_recursive(child, optimizer)


def get_optimiser(
    model,
    base_model_embedding_size=None,
    optimiser=AdamW,
    lr=1e-3,
    weight_decay=1e-2,
    eps=1e-8,
    exclude_keywords=EXCLUDE_FROM_WEIGHT_DECAY,
):
    """
    Set up an optimiser for a transformer model, excluding appropriate submodules
        from weight decay and scaling up weight decay for tensors that have scaled
        from `d_base_model`
    """

    weight_decay_exclude_params = []
    weight_decay_exclude_names = []

    for keyword in exclude_keywords:
        weight_decay_exclude_params += [
            p for name, p in model.named_parameters() if keyword in name.lower()
        ]
        weight_decay_exclude_names += [
            name for name, _ in model.named_parameters() if keyword in name.lower()
        ]

    weight_decay_exclude_params = set(weight_decay_exclude_params)

    if len(weight_decay_exclude_params) > 0:
        warnings.warn(
            "Excluded the following parameters from weight decay based on "
            f"exclude keywords: {set(weight_decay_exclude_names)}",
            stacklevel=2,
        )

    levels = defaultdict(list)

    for _, p in model.named_parameters():

        if (len(p.size()) == 2) and (p not in weight_decay_exclude_params):
            _, in_features = p.size()
            weight_decay_coefficient = math.sqrt(in_features) / math.sqrt(
                base_model_embedding_size
            )
        else:
            weight_decay_coefficient = 0.0

        levels[(lr, weight_decay * weight_decay_coefficient)].append(p)

    parameter_groups = []

    for k, v in levels.items():
        group_learning_rate, group_weight_decay = k
        parameter_groups.append(
            {"params": v, "weight_decay": group_weight_decay, "lr": group_learning_rate}
        )

    return optimiser(
        parameter_groups,
        weight_decay=weight_decay,
        lr=lr,
        eps=eps,
    )



"""
Based on Smith (2017) https://arxiv.org/abs/1506.01186
"""

from typing import Optional
import copy
import math

from torch.amp import GradScaler

from .cycles import Cycle


class PASS:
    """
    A self-configuring learning rate scheduler
    """

    def __init__(
        self,
        learning_rate_schedule: Cycle,
        model,
        optimiser,
        scaler: Optional[GradScaler] = None,
        range_test: bool = False,
        cool_point_multiplier: float = 1 / 60,
    ):
        """
        If not using range test, we assume the optimiser has the learning rates
            set as desired.
        """

        if type(model).__name__ == "OptimizedModule":
            raise ValueError(
                "PASS received a compiled model. You must initialize PASS with the "
                "raw model *before* calling torch.compile().\n"
                "Correct order:\n"
                "  scheduler = PASS(..., model, ...)\n"
                "  model = torch.compile(model)"
            )

        self.model = model
        self.optimiser = optimiser
        self.scaler = scaler

        self.learning_rate_schedule = learning_rate_schedule

        self.range_test = range_test

        self.original_param_groups = copy.deepcopy(optimiser.param_groups)

        self.cool_point_multiplier = cool_point_multiplier

        self.step_count = 0

        self.update_learning_rates()

        self.range_test_results = []

        if self.range_test:
            self.start_range_test()  # sets LR to 1E-7

    @property
    def lr(self):
        """
        Return first lr from self.optimiser.param_groups
            (this is used in learning rate range tests, in which case we can
            assume they are all the same!)
        """
        for group in self.optimiser.param_groups:
            return group["lr"]

    @property
    def in_range_test(self):
        if not self.range_test:
            return False
        elif (len(self.range_test_results) == 0) or (
            not math.isnan(self.range_test_results[-1][1])
        ):
            return True
        else:
            return False

    @property
    def trained(self):
        if not self.range_test:
            return True
        elif math.isnan(self.range_test_results[-1][1]):
            return True
        else:
            return False

    @property
    def finished(self):
        return self.step_count >= len(self.learning_rate_schedule) - 1

    def clean_model_state_dict(self, state_dict):
        new_state_dict = {}
        for k, v in state_dict.items():
            if k.startswith("_orig_mod."):
                new_state_dict[k[10:]] = v  # Strip the "_orig_mod." prefix (length 10)
            else:
                new_state_dict[k] = v
        return new_state_dict

    def _current_states(self):
        saved_states = {
            "model": copy.deepcopy(
                self.clean_model_state_dict(self.model.state_dict())
            ),
            "optimiser": copy.deepcopy(self.optimiser.state_dict()),
        }
        if self.scaler is not None:
            saved_states["scaler"] = copy.deepcopy(self.scaler.state_dict())
        return saved_states

    def save_states(self):
        self.saved_states = self._current_states()

    def load_states(self, states: dict):
        self.model.load_state_dict(self.clean_model_state_dict(states["model"]))
        self.optimiser.load_state_dict(states["optimiser"])
        if self.scaler is not None:
            self.scaler.load_state_dict(states["scaler"])
        self.update_learning_rates()

    def recover_states(self):
        self.load_states(self.saved_states)

    def state_dict(self):
        return {
            "model": self.model.state_dict(),
            "optimiser": self.optimiser.state_dict(),
            "scaler": self.scaler.state_dict() if self.scaler is not None else None,
            "range_test": self.range_test,
            "range_test_results": self.range_test_results,
            "original_param_groups": self.original_param_groups,
            "cool_point_multiplier": self.cool_point_multiplier,
            "step_count": self.step_count,
        }

    def load_state_dict(self, state_dict):
        self.load_states(state_dict)
        self.range_test = state_dict["range_test"]
        self.range_test_results = state_dict["range_test_results"]
        self.original_param_groups = state_dict["original_param_groups"]
        self.cool_point_multiplier = state_dict["cool_point_multiplier"]
        self.step_count = state_dict["step_count"]

    @property
    def _schedule_multiplier(self):
        return self.learning_rate_schedule(
            min(self.step_count, self.learning_rate_schedule.total_steps)
        )

    def set_all_lr(self, lr):
        for group in self.optimiser.param_groups:
            group["lr"] = lr

    def start_range_test(self):
        self.range_test = True
        self.range_test_results = []
        self.save_states()
        self.set_all_lr(1e-7)

    def scale_all_lr(self, scaling_factor):
        self.set_all_lr(self.lr * scaling_factor)

    def end_range_test(self):
        self.recover_states()
        self.update_learning_rates()

    def _smoothed_range_test(self, range_test_results):
        range_test_results = sorted(range_test_results, key=lambda x: x[0])
        learning_rates = [t[0] for t in range_test_results]
        losses = [t[1] for t in range_test_results]
        losses = losses[:-1] + [10 * max(losses)]
        return list(zip(learning_rates, losses, strict=True))

    def _plot_range_test(self, range_test_results):
        """
        Returns a tuple with x values (learning rates) and y values (losses)
            which can then be passed to e.g. pyplot. We recommend presenting
            the plot with a logarithmic x axis.
        """
        range_test_results = sorted(range_test_results, key=lambda x: x[0])
        learning_rates = [t[0] for t in range_test_results]
        losses = [t[1] for t in range_test_results]
        return learning_rates, losses

    def _apply_range_test_result(self):
        """
        ...
        """
        range_test_results = self._smoothed_range_test(self.range_test_results)
        minimum = min(range_test_results, key=lambda x: x[1])
        points_left_of_min = [r for r in range_test_results if r[0] < minimum[0]]
        max_left_of_min = max(points_left_of_min, key=lambda x: x[1])
        difference = max_left_of_min[1] - minimum[1]
        max_lr = None
        for p in sorted(points_left_of_min, key=lambda x: x[0]):
            if (max_lr is None) and (p[1] < minimum[1] + 0.2 * difference):
                max_lr = p[0]
            else:
                continue
        self.set_all_lr(max_lr)
        self.original_param_groups = copy.deepcopy(self.optimiser.param_groups)
        print("High LR", max_lr)

    def update_learning_rates(self):
        if not self.finished:
            for original, current in zip(
                self.original_param_groups, self.optimiser.param_groups, strict=True
            ):
                base_lr = original["lr"]
                min_lr = base_lr * self.cool_point_multiplier
                current_lr = min_lr + (base_lr - min_lr) * self._schedule_multiplier
                current["lr"] = current_lr

    def _append_to_range_test(self, loss_item: float):

        lr = self.lr

        self.range_test_results.append((lr, loss_item))

        if math.isnan(loss_item) or (lr >= 1.0):
            self._apply_range_test_result()
            self.end_range_test()
        else:
            # Continue range test, step up learning rate
            self.scale_all_lr(1.05)

    def step(self, loss_item: Optional[float] = None):
        """
        This function manages the process of
            * Doing an initial range test
            * Training for one microcycle using the learning rates from the
                  initial range test ("burn in")
            * Doing a second range test to set the learning rate schedule for
                  the rest of training
            * Updating learning rates during training according to the macrocycle
        """
        if self.in_range_test:  # True at init unless self.range_test = False
            if not isinstance(loss_item, float):
                raise ValueError(
                    "When using range test functionality, "
                    "`step()` expects a loss item."
                )
            self._append_to_range_test(loss_item)
        elif self.trained and not self.finished:
            self.step_count += 1
            self.update_learning_rates()
        else:
            pass
